{***********************************************************}
{   This file is part of the DORB project.                  }
{                                                           }
{   See the file License, included in this distribution,    }
{   for details about the copyright.                        }
{                                                           }
{   Send comments and/or bug reports to: vic@inec.ru        }
{***********************************************************}
unit naming_impl;

interface

uses
  orb_int,orb,req_int,code_int,imr,imr_int,env_int,except_int,types,
  exceptions,stdstat,std_seq,classes,cosnaming_int, cosnaming, poa_int;

type

  TNamingContext_impl = class(TNamingContext_serv)
  private
    FTable : TStringList;
  protected
    procedure bind(const n: TName; const obj: IORBObject); override;
    procedure rebind(const n: TName; const obj: IORBObject); override;
    procedure bind_context(const n: TName; const nc: INamingContext); override;
    procedure rebind_context(const n: TName; const nc: INamingContext); override;
    function resolve(const n: TName): IORBObject; override;
    procedure unbind(const n: TName); override;
    function new_context: INamingContext; override;
    function bind_new_context(const n: TName): INamingContext; override;
    procedure _destroy; override;
    procedure list(const how_many: _ulong; out bl: TBindingList; out bi: IBindingIterator); override;
  public
    constructor Create(const orb: IORB; const poa: IPOA); overload;
    constructor Create(); overload;
    destructor Destroy; override;
    procedure do_bind(const n: TName; const obj: IORBObject; t: TBindingType; rebind: Boolean);
  end;

  TBindingIterator_impl = class(TBindingIterator_serv)
  private
    FORB : IORB;
    FList : TBindingList;
    FIndex : integer;
  protected
    function next_one(out b: TBinding): boolean; override;
    function next_n(const how_many: _ulong; out bl: TBindingList): boolean; override;
    procedure _destroy; override;
  public
    constructor Create(const orb: IORB; lst: TBindingList);
  end;

  TLocalBinding = class
    FBindingType : TBindingType;
    FObject: IORBObject;
  end;

implementation

uses
  poa;

var
  Contexts : IInterfaceList;
  _poa : IPOA;
  _orb : IORB;
  root : INamingContext;

//***********************************************************
// NamingContext_impl
//***********************************************************
constructor TNamingContext_impl.Create(const orb: IORB; const poa: IPOA);
begin
  inherited Create();
  _orb := orb;
  _poa := poa;
  root := self;
  FTable := TStringList.Create;
  FTable.Sort;
end;

constructor TNamingContext_impl.Create();
begin
  inherited;
  FTable := TStringList.Create;
  FTable.Sort;
end;

destructor TNamingContext_impl.Destroy;
begin
  FTable.Free;
  if root = (self as INamingContext) then
    begin
      root := nil;
      _poa := nil;
      _orb := nil;
    end;
  inherited Destroy;
end;

procedure TNamingContext_impl.do_bind(const n: TName; const obj: IORBObject; t: TBindingType; rebind: Boolean);
var
  i : integer;
  b : TLocalBinding;
  nc,nextnc : INamingContext;
  name: TName;
begin
  name := nil;
  if Length(n) = 0 then
    raise TNamingContext_InvalidName.Create();
  i := FTable.IndexOf(n[0].id+'/'+n[0].kind);
  if length(n) = 1 then
    begin
      if (i >= 0) and not rebind then
        raise TNamingContext_AlreadyBound.Create()
      else
        begin
          if i < 0 then
            b := TLocalBinding.Create
          else
            b := TLocalBinding(FTable.Objects[i]);
          b.FBindingType := t;
          b.FObject := obj;
          if i < 0 then
            FTable.AddObject(n[0].id+'/'+n[0].kind,b);
        end;
    end
  else
    begin
      if i < 0 then   // not found
        raise TNamingContext_NotFound.Create(missing_node,n);
      b := TLocalBinding(FTable.Objects[i]);
      if b.FBindingType <> ncontext then // context required
        raise TNamingContext_CannotProceed.Create;
      nc := TNamingContext._narrow(b.FObject);
      assert(nc<>nil);
      name := copy(n,1,Length(n)-1);
      case t of
        nobject : if rebind then
            nc.rebind(name,obj)
          else
            nc.bind(name,obj);
        ncontext : begin
            nextnc := TNamingContext._narrow(obj);
            assert(nextnc <> nil);
            if rebind then
              nc.rebind_context(name,nextnc)
            else
              nc.bind_context(name,nextnc);
          end;
      end;
    end;
end;

procedure TNamingContext_impl.bind(const n: TName; const obj: IORBObject);
begin
  do_bind(n,obj,nobject,false);
end;

procedure TNamingContext_impl.rebind(const n: TName; const obj: IORBObject);
begin
  do_bind(n,obj,nobject,true);
end;

procedure TNamingContext_impl.bind_context(const n: TName; const nc: INamingContext);
begin
  do_bind(n,nc as IORBObject,ncontext,false);
end;

procedure TNamingContext_impl.rebind_context(const n: TName; const nc: INamingContext);
begin
  do_bind(n,nc as IORBObject,ncontext,true);
end;

function TNamingContext_impl.resolve(const n: TName): IORBObject;
var
  i : integer;
  b: TLocalBinding;
  nc : INamingContext;
  name: TName;
begin
  name := nil;
  if Length(n) = 0 then
    raise TNamingContext_InvalidName.Create();
  i := FTable.IndexOf(n[0].id+'/'+n[0].kind);
  if i < 0 then   // not found
    raise TNamingContext_NotFound.Create(missing_node,n);
  b := TLocalBinding(FTable.Objects[i]);
  if Length(n) = 1 then
    result := b.FObject
  else
    begin
      if b.FBindingType <> ncontext then // context required
        raise TNamingContext_CannotProceed.Create;
      nc := TNamingContext._narrow(b.FObject);
      assert(nc<>nil);
      name := copy(n,1,Length(n)-1);
      result := nc.resolve(name);
    end;
end;

procedure TNamingContext_impl.unbind(const n: TName);
var
  i : integer;
  b: TLocalBinding;
  nc : INamingContext;
  name: TName;
begin
  name := nil;
  if Length(n) = 0 then
    raise TNamingContext_InvalidName.Create();
  i := FTable.IndexOf(n[0].id+'/'+n[0].kind);
  if i < 0 then   // not found
    raise TNamingContext_NotFound.Create(missing_node,n);
  b := TLocalBinding(FTable.Objects[i]);
  if Length(n) = 1 then
    begin
      FTable.Objects[i].Free;
      FTable.Delete(i);
    end
  else
    begin
      if b.FBindingType <> ncontext then // context required
        raise TNamingContext_CannotProceed.Create;
      nc := TNamingContext._narrow(b.FObject);
      assert(nc<>nil);
      name := copy(n,1,Length(n)-1);
      nc.unbind(name);
    end;
end;

function TNamingContext_impl.new_context: INamingContext;
var
  nnc : TNamingContext_impl;
  oid: string;
  obj : IORBObject;
begin
  nnc := TNamingContext_impl.Create();
  oid := FPOA.activate_object(nnc);
  obj := FPOA.id_to_reference(oid);
  Contexts.Add(nnc);
  result := TNamingContext._narrow(obj);
end;

function TNamingContext_impl.bind_new_context(const n: TName): INamingContext;
begin
  result := new_context;
  bind_context(n,result);
end;

procedure TNamingContext_impl._destroy;
var
  pc: IPOACurrent;
  obj: IORBObject;
begin
  if FTable.Count > 0 then
    raise TNamingContext_NotEmpty.Create();
  obj := _orb.resolve_initial_reference('POACurrent');
  pc := TPOACurrent._narrow(obj);
  pc.get_POA.deactivate_object(pc.get_object_id);
  Contexts.Remove(self);
end;

procedure TNamingContext_impl.list(const how_many: _ulong; out bl: TBindingList; out bi: IBindingIterator);
var
  num: integer;
  i,j,last,ind : integer;
  rest_items: TBindingList;
begin
  num := how_many;
  if num > FTable.Count then num := FTable.Count;
  setLength(bl,num);
  for i := 0 to Pred(num) do
    begin
      ind := Pos('/',FTable[i]);
      setLength(bl[i].binding_name,1);
      bl[i].binding_name[0].id := copy(FTable[i],1,ind-1);
      bl[i].binding_name[0].kind := copy(FTable[i],ind+1,10000);
      bl[i].binding_type := TLocalBinding(FTable.Objects[i]).FBindingType;
    end;
  if integer(how_many) < FTable.Count then
    begin
      last := num;
      num := FTable.Count - integer(how_many);
      setLength(rest_items,num);
      for j := last to Pred(num) do
        begin
          ind := Pos('/',FTable[j]);
          setLength(rest_items[j].binding_name,1);
          rest_items[j].binding_name[0].id := copy(FTable[j],1,ind-1);
          rest_items[j].binding_name[0].kind := copy(FTable[j],ind+1,10000);
          rest_items[j].binding_type := TLocalBinding(FTable.Objects[j]).FBindingType;
        end;
      bi := TBindingIterator_impl.Create(_orb,rest_items)._this;
    end;
end;

//***********************************************************
// BindingIterator_impl
//***********************************************************
constructor TBindingIterator_impl.Create(const orb: IORB;
  lst: TBindingList);
begin
  inherited Create();
  FIndex := 0;
  FORB := orb;
  FList := lst;
end;

function TBindingIterator_impl.next_one(out b: TBinding): boolean;
begin
  result := false;
  if FIndex < Length(FList) then
    begin
      result := true;
      b := FList[FIndex];
      inc(FIndex);
    end;
end;

function TBindingIterator_impl.next_n(const how_many: _ulong; out bl: TBindingList): boolean;
var
  num : integer;
begin
  result := false;
  if FIndex < Length(FList) then
    begin
      num := how_many;
      if Length(FList) < (FIndex+num) then num := Length(FList) - FIndex;
      result := true;
      bl := copy(FList,FIndex,num);
      inc(FIndex,num);
    end
  else
    setLength(bl,0);
end;

procedure TBindingIterator_impl._destroy;
var
  pc: IPOACurrent;
begin
  pc := TPOACurrent._narrow(_orb.resolve_initial_reference('POACurrent'));
  pc.get_POA.deactivate_object(pc.get_object_id);
end;

initialization
  Contexts := TInterfaceList.Create;
end.
